-- get the addon namespace
local addon, ns = ...
-- get the config values
local cfg = ns.cfg
-- holder for some lib functions
local lib = _G["HEDD_lib"] or CreateFrame("Frame","HEDD_lib")
ns.lib = lib

lib.fullcdtime = function(spellid)
	Hedd_Tooltip = _G["Hedd_Tooltip"] or CreateFrame('GameTooltip', 'Hedd_Tooltip', UIParent, 'GameTooltipTemplate')
	Hedd_Tooltip:SetOwner(UIParent, 'ANCHOR_NONE')
	Hedd_Tooltip:SetSpellByID(spellid)
	text=Hedd_TooltipTextRight3:GetText()
	Hedd_Tooltip:Hide()
	return hedlib.tofloat(text)
end

lib.UpdateSpellTT = function(spell)
	if spell and cfg.spells[spell] then
		if cfg.spells[spell].tt.fullcd then
			cfg.spells[spell].fullcd=hedlib.tofloat(_G[cfg.spells[spell].tt.fullcd]:GetText() or " ")
		else
			cfg.spells[spell].fullcd=0
		end

		if cfg.spells[spell].type=="cd" then
			cfg.spells[spell].cost=0
			return
		end
		
		if cfg.spells[spell].cost_real then
			cfg.spells[spell].cost=cfg.spells[spell].cost_real
		else
			if cfg.spells[spell].tt.cost then
				cfg.spells[spell].cost=hedlib.tofloat(_G[cfg.spells[spell].tt.cost]:GetText() or " ")
			else
				cfg.spells[spell].cost=0
			end
		end
	end
end

local function gsi_inner(...)
	if ... == "" then
		return nil
	else
		return ...
	end
end

function lib.Hedd_GSI(...) -- name, rank, icon, castTime, minRange, maxRange
	return gsi_inner(GetSpellInfo(...))
end

function Hedd_GSI(...)
	return lib.Hedd_GSI(...)
end

lib.AddSpell = function(spell,ids,nousecheck,powerCost_real) --powertype,
	cfg.spells[spell]=nil
	--[[powertype=(powertype==true) and "alt" or powertype
	powertype=powertype or "power"]]
	nousecheck=nousecheck or nil
	powerCost_real = powerCost_real or nil
	for index,id in ipairs(ids) do
		if IsPlayerSpell(id) then
			cfg.spells[spell]={}
			cfg.spells[spell].id=id
			cfg.spells[spell].f = _G["HEDD_"..spell] or CreateFrame("Frame","HEDD_"..spell,Heddframe)
			cfg.spells[spell].f:SetAllPoints(Heddframe)
			cfg.spells[spell].name, _, cfg.spells[spell].icon, cfg.spells[spell].castingTime, cfg.spells[spell].minRange, cfg.spells[spell].maxRange = lib.Hedd_GSI(id)
			lib.SetSpellIcon(spell)
			cfg.sp_conv[cfg.spells[spell].name]=spell
			cfg.id2spell[cfg.spells[spell].id]=spell
			cfg.spells[spell].noupdate=noupdate or nil
			cfg.spells[spell].cd = 0
			cfg.spells[spell].guid = 0
			cfg.spells[spell].lastcast=0
			cfg.spells[spell].start = 0
			cfg.spells[spell].isUsable = 1
			cfg.spells[spell].notEnoughMana = nil
			cfg.spells[spell].cost = 0
			--cfg.spells[spell].powerType = 0
			cfg.spells[spell].channel = 0
			cfg.spells[spell].buff=0
			--cfg.spells[spell].type=powertype
			cfg.spells[spell].nousecheck=nousecheck
			cfg.spells[spell].currentCharges=1
			cfg.spells[spell].maxCharges=1
			if powerCost_real then cfg.spells[spell].cost_real=powerCost_real end
			cfg.spells[spell].tt=_G["Hed_Tooltip_"..cfg.spells[spell].id] or CreateFrame('GameTooltip', "Hed_Tooltip_"..cfg.spells[spell].id, nil, 'GameTooltipTemplate')
			cfg.spells[spell].tt:SetOwner(UIParent, 'ANCHOR_NONE')
			cfg.spells[spell].tt:SetSpellByID(cfg.spells[spell].id)
			cfg.spells[spell].tt.id=id
			cfg.spells[spell].tt.fullcd=hedlib.ScanTooltip(cfg.spells[spell].tt,cfg.local_cd,spell)
			cfg.spells[spell].tt.cost=hedlib.ScanTooltip(cfg.spells[spell].tt,cfg.Power.local_type,spell)
			if cfg.spells[spell].tt.cost then
				cfg.spells[spell].type="power"
			else
				if cfg.AltPower.local_type then
					cfg.spells[spell].tt.cost=hedlib.ScanTooltip(cfg.spells[spell].tt,cfg.AltPower.local_type,spell)
					if cfg.spells[spell].tt.cost then
						cfg.spells[spell].type="alt"
					else
						cfg.spells[spell].type="cd"
					end
				else
					cfg.spells[spell].type="cd"
				end
			end
			cfg.spells[spell].tt:Hide()
			lib.UpdateSpell(spell)
			cfg.Update=true
			return true
		end
	end
	return false
end

lib.SetSpellCost = function(spell,powercost,powertype)
	if cfg.spells[spell] then
		cfg.spells[spell].cost_real=powercost
		cfg.spells[spell].type=powertype
	end
end

local icon
lib.GetSpellIcon = function(spell)
	if cfg.spells[spell] then return cfg.spells[spell].icon end
	icon=(select(3,GetSpellInfo(spell)))
	return icon
end

lib.GetSpellCharges = function(spell)
	if not cfg.spells[spell] then return 0 end
	if cfg.spells[spell].currentCharges then return cfg.spells[spell].currentCharges end
	if cfg.spells[spell].isUsable and cfg.spells[spell].cd==0 then return 1 end
	return 0
end

lib.GetSpellMaxCharges = function(spell)
	if not cfg.spells[spell] then return 0 end
	if cfg.spells[spell].currentCharges then return cfg.spells[spell].maxCharges end
	return 1
end

lib.SpellInTime = function (spell , mint, maxt)
	if not cfg.spells[spell] then return nil end
	if cfg.spells[spell].lastcast >= mint and cfg.spells[spell].lastcast <=maxt then
		return true
	else
		return nil
	end
end

lib.SpellName2Spell = function(spellName)
	return cfg.sp_conv[spellName] or nil
end

lib.SpellId2Spell = function(id)
	return cfg.id2spell[id] or nil
end

lib.KnownSpell = function(spell)
	if cfg.spells[spell] then return true end
	return nil
end

lib.UpdateSpellCharges = function(spell)
	if cfg.spells[spell] then
	cfg.spells[spell].currentCharges, cfg.spells[spell].maxCharges = GetSpellCharges(cfg.spells[spell].id)
	end
end

lib.UpdateAllSpells = function(powertype)
	for index,name in pairs(cfg.spells) do
		if cfg.spells[index] and not cfg.spells[index].noupdate and ((powertype and cfg.spells[index].type==powertype) or not powertype) then
			lib.UpdateSpell(index)
		end
	end
end

spells_old={}
lib.UpdateSpell = function(spell,now)
	spell=cfg.spells[spell] and spell or cfg.id2spell[spell]
	if cfg.spells[spell] then
		--[[if cfg.spells_updating[spell] then
			print(spell.." is already updating!")
		end
		cfg.spells_updating[spell]=true]]
		hedlib.shallowCopy(cfg.spells[spell],spells_old)
		if now then cfg.spells[spell].lastcast=GetTime() end
		cfg.spells[spell].start, cfg.spells[spell].cd, senabled = GetSpellCooldown(cfg.spells[spell].id)
		if not (cfg.spells[spell].start>0 and cfg.spells[spell].cd>0) then -- and senabled
			cfg.spells[spell].cd=0
		end

		cfg.spells[spell].isUsable,cfg.spells[spell].notEnoughMana = IsUsableSpell(cfg.spells[spell].id)
		cfg.spells[spell].castingTime = select(4,lib.Hedd_GSI(cfg.spells[spell].id))/1000
		if cfg.spells[spell].castingTime<0 then cfg.spells[spell].castingTime=0 end
		cfg.spells[spell].tt=_G["Hed_Tooltip_"..cfg.spells[spell].id] or CreateFrame('GameTooltip', "Hed_Tooltip_"..cfg.spells[spell].id, nil, 'GameTooltipTemplate')
		cfg.spells[spell].tt:SetOwner(UIParent, 'ANCHOR_NONE')
		cfg.spells[spell].tt:SetSpellByID(cfg.spells[spell].id)
		lib.UpdateSpellTT(spell)
		cfg.spells[spell].tt:Hide()
		lib.UpdateSpellCharges(spell)

		if not hedlib.ArrayNotChanged(cfg.spells[spell],spells_old) or now then
			if cfg.spells[spell].OnUpdate then cfg.spells[spell].OnUpdate() end
			if cfg.debugspells then print("Updating "..spell) end
			cfg.Update=true
			--return true
		end
		--cfg.spells_updating[spell]=false
	end
	--return nil
end

totems_old={}
lib.UpdateTotem = function(totemtype)
	totemtype = totemtype or 1
	if not cfg.totems then cfg.totems={} end
	if not cfg.totems[totemtype] then
		cfg.totems[totemtype]={}
		cfg.totems[totemtype].haveTotem, cfg.totems[totemtype].totemName, cfg.totems[totemtype].startTime, cfg.totems[totemtype].duration, cfg.totems[totemtype].icon = GetTotemInfo(totemtype)
		cfg.Update=true
		return true
	end
	hedlib.shallowCopy(cfg.totems[totemtype],totems_old)
	cfg.totems[totemtype].haveTotem,cfg.totems[totemtype].totemName,cfg.totems[totemtype].startTime,cfg.totems[totemtype].duration, cfg.totems[totemtype].icon = GetTotemInfo(totemtype)
	
	if not hedlib.ArrayNotChanged(cfg.totems[totemtype],totems_old) then
		cfg.Update=true
		return true
	end
	return nil
end

lib.MyTotem = function(totemtype)
	totemtype=totemtype or 1
	if not cfg.totems then lib.UpdateTotem(totemtype) end
	if not cfg.totems[totemtype].haveTotem then return nil end
	return cfg.totems[totemtype].icon
end

lib.GetTotem = function(totemtype)
	totemtype=totemtype or 1
	if not cfg.totems then lib.UpdateTotem(totemtype) end
	if not cfg.totems[totemtype].haveTotem then return 0 end
	
	tl_totem = cfg.totems[totemtype].duration - (GetTime()-cfg.totems[totemtype].startTime)
	tl_totem = (tl_totem>0) and tl_totem or 0
	
	return tl_totem
end

lib.HaveTotem = function(spell,totemtype)
	totemtype=totemtype or 1
	if cfg.spells[spell] and cfg.totems[totemtype] and lib.GetTotem(totemtype)>0 and cfg.spells[spell].name==cfg.totems[totemtype].totemName then return true end
	return nil
end

local i_spell,noprio,prio_count,i_case,n_case
lib.SetPriority = function()
	cfg.FindPriority=true
	for i_spell in pairs(cfg.spells) do
		lib.HideFrame(cfg.spells[i_spell].f)
	end

	cfg.timeleft = 0
	cfg.ctimeleft = cfg.maxmintime
	noprio=true
	prio_count=0
	while noprio do
		prio_count=prio_count+1
		for i_case, n_case in pairs(cfg.plist) do
			if cfg.case[n_case] then
				if cfg.case[n_case]() then
					noprio=nil
					break
				end
			else
				if lib.SimpleCDCheck(n_case) then
					noprio=nil
					break
				end
			end
		end
		if cfg.ctimeleft==cfg.maxmintime then
			lib.UpdateBar("aa")
			noprio=nil
		end
		if prio_count>cfg.numcase then
			lib.UpdateBar("aa")
			noprio=nil
		end
		cfg.timeleft=cfg.ctimeleft
	end
	Heddframe.bar.force=true
end

local tl_spell
lib.SimpleCDCheck = function (spell, Wait,nousecheck,nothing2do) -- nomanacheck, nousecheck, useGCD
	if not cfg.spells[spell] then return nil end
	if HeddDB.CD[cfg.spells[spell].id] and not HeddDB.CD[cfg.spells[spell].id].enabled then return nil end
	if cfg.spells[spell].notEnoughMana and cfg.spells[spell].type=="alt" then return nil end
	Wait=Wait or 0
	nothing2do=nothing2do or 0
	nousecheck=nousecheck or nil
	if (cfg.spells[spell].type=="power" and (cfg.spells[spell].isUsable or cfg.spells[spell].notEnoughMana))
	or ((cfg.spells[spell].type=="alt" or cfg.spells[spell].type=="cd") and cfg.spells[spell].isUsable)
	or nousecheck  then --or cfg.spells[spell].nousecheck
		tl_spell = math.max(0,lib.GetSpellCD(spell),Wait)

		if nothing2do==0 then
			if tl_spell<=(cfg.timeleft+cfg.react) then --tl_spell==0 or
				lib.UpdateBar(spell,tl_spell)
				cfg.ctimeleft=tl_spell
				return true
			else
				if cfg.ctimeleft>tl_spell then
					cfg.ctimeleft=tl_spell
				end
			end
		else
			if (cfg.timeleft+cfg.react)>=(tl_spell+nothing2do) then
				lib.UpdateBar(spell,tl_spell)
				cfg.ctimeleft=tl_spell
				return true
			else 
				if cfg.ctimeleft>(tl_spell+nothing2do) then
					cfg.ctimeleft=(tl_spell+nothing2do)
				end
			end
		end
	end
	return nil
end

lib.isSpellUsable=function(spell)
	if cfg.spells[spell] and cfg.spells[spell].isUsable then
		return true
	else
		return false
	end
end

local tl_spell_cd
--[[lib.GetSpellCD = function (spell,nopower)
	--tl_spell_cd=9999
	spell = (spell=="gcd") and cfg.gcd or spell
	if cfg.spells[spell] then
		tl_spell_cd=math.max(0,(cfg.spells[spell].cd - (GetTime() - cfg.spells[spell].start)))
		if not nopower and cfg.spells[spell].type=="power" and cfg.Power.regen>0 then --and cfg.Power.real<cfg.spells[spell].cost
			tl_spell_cd=math.max(tl_spell_cd,((cfg.spells[spell].cost-cfg.Power.real)/cfg.Power.regen))
		end
		return math.max(tl_spell_cd,lib.SpellCastingLeft())
	end
	return 9999
end]]

lib.GetSpellCD = function (spell)
	spell = (spell=="gcd") and cfg.gcd or spell
	if cfg.spells[spell] then
		tl_spell_cd=math.max(0,(cfg.spells[spell].cd - (GetTime() - cfg.spells[spell].start)))
		if cfg.spells[spell].type=="power" then
			tl_spell_cd=math.max(tl_spell_cd,lib.Time2Power(cfg.spells[spell].cost))
		else
			tl_spell_cd=math.max(tl_spell_cd,lib.SpellCastingLeft())
		end
		return tl_spell_cd
	else
		return 9999
	end
end


local nextspell=nil
local nextspell_i=nil
lib.GetSpellCDNext = function()
	nextspell=nil
	for index,name in pairs(cfg.spells) do
		if cfg.spells[index] and not cfg.spells[index].noupdate then 
			nextspell_i=lib.GetSpellCD(index)
			if nextspell_i>0 then
				nextspell=nextspell or nextspell_i
				nextspell=math.min(nextspell,nextspell_i)
			end
		end
	end
	if nextspell then return (GetTime()+nextspell) end
	return 0
end

lib.GetSpellFullCD = function (spell)
	if spell=="gcd" then spell=cfg.gcd end
	if cfg.spells[spell] then
		return cfg.spells[spell].fullcd
	end
	return 0
end

--[[lib.GetSpellCDFull = function (spell)
	if cfg.spells[spell] then
		return cfg.spells[spell].cd
	end
	return 0
end]]

lib.GetSpellCost = function (spell)
	if cfg.spells[spell] and (cfg.spells[spell].type=="power" or cfg.spells[spell].type=="alt") then
		return cfg.spells[spell].cost
	end
	return 0
end

lib.GetSpellCostbyName = function(name)
	if name and cfg.sp_conv[name] then
		return cfg.spells[cfg.sp_conv[name]].cost
	else
		return 0
	end
end

lib.GetSpellIdbyName = function(name)
	if name and cfg.sp_conv[name] then
		return cfg.spells[cfg.sp_conv[name]].id
	else
		return 0
	end
end

lib.GetSpellName = function (spell)
	if cfg.spells[spell] then
		return cfg.spells[spell].name
	end
	return "nospell"
end

lib.GetSpellId = function (spell)
	if cfg.spells[spell] then
		return cfg.spells[spell].id
	end
	return "nospell"
end

lib.GetSpellCT = function (spell)
	if spell=="gcd" then spell=cfg.gcd end
	
	if cfg.spells[spell] and cfg.spells[spell].castingTime then
		return cfg.spells[spell].castingTime
	end
	return 0
end

lib.Time2Wait=function(spell)
	return math.max(lib.GetSpellCD(spell),lib.SpellCastingLeft())
end


lib.SetSpellFunction = function(spell,event,func)
	if cfg.spells[spell] then
		if event=="OnUpdate" then
			cfg.spells[spell].OnUpdate=func
		end
		if event=="OnEnd" then
			cfg.spells[spell].OnEnd=func
		end
	end
end


lib.PrintSpell=function(spell)
	if cfg.spells[spell] then
		tmp2=spell
		tmp2=tmp2.." "..cfg.spells[spell].name
		tmp2=tmp2.." id="..cfg.spells[spell].id
		tmp2=tmp2.." ct="..cfg.spells[spell].castingTime
		tmp2=tmp2.." cd="..cfg.spells[spell].cd
		tmp2=tmp2.." fullcd="..cfg.spells[spell].fullcd
		tmp2=tmp2.." start="..cfg.spells[spell].start
		tmp2=tmp2.." use="..tostring(cfg.spells[spell].isUsable)
		tmp2=tmp2.." nomana="..tostring(cfg.spells[spell].notEnoughMana)
		tmp2=tmp2.." cost="..cfg.spells[spell].cost
		tmp2=tmp2.." powertype="..tostring(cfg.spells[spell].type)
		print(tmp2)
	else
		print(spell.." not found!")
	end
end

lib.PrintAllSpells = function()
	for index,name in pairs(cfg.spells) do
		lib.PrintSpell(index)
	end
end

local isUsable,notEnoughMana
lib.printIsUsable=function(spellid)
	isUsable,notEnoughMana = IsUsableSpell(spellid)
	print(spellid.." usable="..tostring(isUsable).." notEnoughMana="..tostring(notEnoughMana))
end